/*******************************************************************************
 * Copyright (c) 2008 Plant Bioinformatics Group, IPK Gatersleben
 *******************************************************************************/
package de.ipk_gatersleben.ag_pbi.fluxbalance.reconstruction;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import org.AttributeHelper;
import org.ErrorMsg;
import org.OpenFileDialogService;
import org.StringManipulationTools;
import org.graffiti.editor.MainFrame;
import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;

import de.ipk_gatersleben.ag_pbi.fluxbalance.startsim.AbstractStartCalculationAlgorithm;
import de.ipk_gatersleben.ag_pbi.fluxbalance.startsim.FBAHelper;

/**
 * @author Eva
 */
public class CreateCmatrix extends AbstractStartCalculationAlgorithm {
	
	public String getName() {
		return "<html>" +
							"<font color=\"#0101DF\">Create C-matrix file</font>";
		
	}
	
	@Override
	public String getDescription() {
		return "<html>" +
							"<b>With this command you may create the C-matrix file required<br>" +
							"for model analysis.<br>" +
							"<br>";
	}
	
	public void execute() {
		File file = CreateMetaboliteFile.askUserForMetaboliteFile(new File(CreateMetaboliteFile.getMetaboliteFilenameFromGraph(graph)));
		if (file != null) {
			ArrayList<String> unblncrcts = doCreateMatrix(file, graph);
			
			// MainFrame.showMessage("C-matrix file created", MessageType.INFO);
			if (unblncrcts == null)
				return;
			if (unblncrcts.size() <= 0)
				MainFrame.showMessageDialog("C-matrix file created successfully", "Information");
			else {
				// MainFrame.showMessage("C-matrix: the following reactions are unbalanced: "+rcts+" Please check!", MessageType.INFO);
				
				MainFrame.showMessageDialog("<html>" +
								"C-matrix file created, but the following reactions are unbalanced:<br>" + getUnbalancedWarningList(unblncrcts)
								, "Information");
				
			}
		}
	}// execute
	
	public static String getUnbalancedWarningList(ArrayList<String> unblncrcts) {
		String rcts = "";
		for (String s : unblncrcts)
			rcts += s + "; ";
		return StringManipulationTools.getWordWrap(rcts.substring(0, rcts.length() - "; ".length()), 100);
	}
	
	public static ArrayList<String> doCreateMatrix(File file, Graph graph) {
		try {
			
			FileReader myFile = null;
			BufferedReader buff = null;
			final List<String> lines = new ArrayList<String>();
			
			try {
				myFile = new FileReader(file);
				buff = new BufferedReader(myFile);
				String line;
				
				while ((line = buff.readLine()) != null) {
					lines.add(line);
				}
				
			} catch (IOException e) {
				ErrorMsg.addErrorMessage(e);
			} finally {
				try {
					buff.close();
					myFile.close();
				} catch (IOException e) {
					ErrorMsg.addErrorMessage(e);
				}
			}
			
			// Metabolite-Hash erzeugen
			final HashMap<String, Integer> metabolites = new HashMap<String, Integer>();
			for (final String line : lines) {
				String[] fields = line.split(":");
				
				int valueInt = 0;
				for (int i = 0; i < 2; i++) {
					String key = fields[0];
					String value = fields[1];
					String value2 = value.trim();
					
					if (value2.contains("\t")) {
						value2 = value2.replaceAll("\t", "");
					}
					try {
						valueInt = Integer.parseInt(value2);
					} catch (NumberFormatException e) {
						ErrorMsg.addErrorMessage("Please check Metabolite file: uncorrect C-atom completion.");
					}
					metabolites.put(key, valueInt);
				}
			}
			
			ArrayList<String> unblncrcts = createMatrix(metabolites, graph);
			
			if (unblncrcts == null)
				return null;
			if (unblncrcts.size() <= 0)
				return new ArrayList<String>();
			else
				return unblncrcts;
			
		} catch (Exception e) {
			ErrorMsg.addErrorMessage(e);
			return null;
		}
	}
	
	private static ArrayList<String> createMatrix(HashMap<String, Integer> metabolites, Graph graph) {
		
		// create C-matrix File
		String absPathfileName = getMatrixFilenameFromGraph(graph);
		
		// create matrices
		ArrayList<String> reactionNames = FBAHelper.getReactionNames(graph.getNodes());
		reactionNames.size();
		int row = reactionNames.size();
		int col = 8; // Anzahl der MatrixSpalten: 4 Edukte + 4 Produkte
		
		HashMap<String, String[][]> result = new HashMap<String, String[][]>();
		result = getRecProdStoiLists(graph.getNodes(), row, col);
		
		if (result == null)
			return null;
		
		// create substitutionMatrix
		String[][] metaboliteMatrix = new String[row][col];
		Integer[][] SubstituteMatrix = new Integer[row][col];
		
		metaboliteMatrix = result.get("met");
		
		for (int i = 0; i < row; i++) {
			for (int j = 0; j < col; j++) {
				String meta = metaboliteMatrix[i][j];
				if (metabolites.containsKey(meta)) {
					int x = metabolites.get(meta);
					SubstituteMatrix[i][j] = x;
				} else {
					SubstituteMatrix[i][j] = 0;
				}
			}
		}
		
		// create resultMatrix
		String[][] StoichCoefMatrix = new String[row][col];
		Integer[][] resultMatrix = new Integer[row][col];
		
		StoichCoefMatrix = result.get("stoich");
		
		ArrayList<String> exceptions = new ArrayList<String>();
		
		for (int i = 0; i < row; i++) {
			for (int j = 0; j < col; j++) {
				String stoich = StoichCoefMatrix[i][j];
				try {
					if (stoich.endsWith(".0"))
						stoich = stoich.substring(0, stoich.length() - ".0".length());
					Integer stoichCoef = Integer.parseInt(StringManipulationTools.removeHTMLtags(stoich));
					Integer subst = SubstituteMatrix[i][j];
					resultMatrix[i][j] = stoichCoef * subst;
				} catch (NumberFormatException e) {
					exceptions.add("\"" + StringManipulationTools.removeHTMLtags(stoich) + "\".");
				}
			}
		}
		if (exceptions.size() > 0) {
			String msg = "Stoichiometric coefficients must be integer numbers. Please change the edge labels";
			for (String e : exceptions)
				msg += "<br>" + e;
			ErrorMsg.addErrorMessage(msg);
			return null;
		}
		// resultMatrix nach unbalancierten Reaktionen durchsuchen
		ArrayList<String> unbalancedRcts = new ArrayList<String>();
		int sumEducts;
		int sumProducts;
		int diff;
		for (int i = 0; i < row; i++) {
			sumEducts = 0;
			sumProducts = 0;
			diff = 0;
			for (int j = 0; j < col; j++) {
				int res = resultMatrix[i][j];
				if (j < 4) {
					sumEducts = sumEducts + res;
				}
				if (j >= 4) {
					sumProducts = sumProducts + res;
				}
			}
			diff = sumEducts - sumProducts;
			if (diff != 0) {
				unbalancedRcts.add(reactionNames.get(i));
			}
		}
		
		// create resultFinalMatrix
		int col2 = 2 * col;
		String[][] resultMatrixFinal = new String[row][col2];
		String[][] VANTEDMatrix = new String[row][col];
		VANTEDMatrix = result.get("Vanted");
		int[] indexI = { 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7 };
		
		for (int i = 0; i < row; i++) {
			for (int j = 0; j < col2; j++) {
				if (j % 2 == 0) {
					resultMatrixFinal[i][j] = VANTEDMatrix[i][indexI[j]] + ";";
				} else {
					resultMatrixFinal[i][j] = (new Integer(resultMatrix[i][indexI[j]])).toString() + ";";
				}
			}
		}
		
		// create Cmatrix File
		createfile(absPathfileName, row, col2, resultMatrixFinal, graph);
		
		return unbalancedRcts;
		
	}
	
	public static String getMatrixFilenameFromGraph(Graph graph) {
		return graph.getName(true).replaceAll(".gml", "Cmatrix.csv");
	}
	
	public static HashMap<String, String[][]> getRecProdStoiLists(Collection<Node> nodes, int row, int col) {
		
		HashMap<String, String[][]> result = new HashMap<String, String[][]>();
		int index = 0;
		String[][] metaboliteMatrix = new String[row][col];
		String[][] StoichCoefMatrix = new String[row][col];
		String[][] VANTEDMatrix = new String[row][col];
		
		for (Node n : nodes) {
			
			if (FBAHelper.isReaction(n)) {
				String rctLabel = AttributeHelper.getLabel(n, "");
				
				ArrayList<String> reacProd = new ArrayList<String>();
				ArrayList<String> stoich = new ArrayList<String>();
				
				for (Edge inEdge : n.getAllInEdges()) {
					Node nS = inEdge.getSource();
					if (nS != null) {
						String reactantLabel = AttributeHelper.getLabel(nS, "");
						reacProd.add(reactantLabel);
						String stoichCoefR = AttributeHelper.getLabel(inEdge, "");
						
						String stoichCoefR2;
						if (stoichCoefR.equals("")) {
							stoichCoefR2 = "1";
						} else {
							stoichCoefR2 = stoichCoefR;
						}
						stoich.add(stoichCoefR2);
					}
				}
				int size = reacProd.size();
				
				if (size > 4) {
					ErrorMsg.addErrorMessage("Reactions containing more than 4 reactants are not allowed, please correct \"" + rctLabel + "\".");
					return null;
				}
				
				while (size < 4) {
					reacProd.add("");
					stoich.add("0");
					size++;
				}
				
				for (Edge outEdge : n.getAllOutEdges()) {
					Node nT = outEdge.getTarget();
					if (nT != null) {
						String productLabel = AttributeHelper.getLabel(nT, "");
						reacProd.add(productLabel);
						String stoichCoefP = AttributeHelper.getLabel(outEdge, "");
						String stoichCoefP2;
						if (stoichCoefP.equals("")) {
							stoichCoefP2 = "1";
						} else {
							stoichCoefP2 = stoichCoefP;
						}
						stoich.add(stoichCoefP2);
					}
				}
				int size1 = reacProd.size();
				
				if (size1 > 8) {
					ErrorMsg.addErrorMessage("Reactions containing more than 4 products are not allowed, please correct \"" + rctLabel + "\".");
					return null;
				}
				
				while (size1 < 8) {
					reacProd.add("");
					stoich.add("0");
					size1++;
				}
				
				String tag = rctLabel;
				String si = "^";
				String tagFinal;
				String name;
				for (int i = 0; i < 8; i++) {
					name = reacProd.get(i);
					metaboliteMatrix[index][i] = name;
					
					StoichCoefMatrix[index][i] = stoich.get(i);
					tagFinal = tag + si + name;
					VANTEDMatrix[index][i] = tagFinal;
				}
				index++;
			}
			
		}// for
		result.put("met", metaboliteMatrix);
		result.put("stoich", StoichCoefMatrix);
		result.put("Vanted", VANTEDMatrix);
		return result;
	}
	
	private static void createfile(String absPathfileName2, int row, int col2, String[][] resultMatrixFinal, Graph graph) {
		
		try {
			BufferedWriter Cmatrixfile = new BufferedWriter(new FileWriter(absPathfileName2));
			for (int i = 0; i < row; i++) {
				for (int j = 0; j < col2; j++) {
					String fin = resultMatrixFinal[i][j];
					Cmatrixfile.write(fin);
				}
				Cmatrixfile.newLine();
			}
			Cmatrixfile.close();
			
		} catch (IOException e1) {
			ErrorMsg.addErrorMessage(e1);
		}
	}
	
	@Override
	protected String getImageFileName() {
		return null;
	}
	
	public static File askUserForCMatrixFile(File activedir) {
		// Einlesen der C-matrix (*.csv)
		if (activedir != null)
			OpenFileDialogService.setActiveDirectoryFrom(activedir);
		File cmatrixf = OpenFileDialogService.getFile(
				new String[] { ".csv" }, "C-matrix File (*.csv)", "Select C-matrix File");
		
		if (cmatrixf == null) {
			MainFrame.showMessageDialog("Please select a C-matrix file (*.csv)!", "Error");
			return null;
		} else
			return cmatrixf;
		
	}
}